@@ -6,8 +6,19 @@ import torch.nn as nn
 from torch.utils.data import DataLoader
 from torch.optim import Optimizer
 import matplotlib.pyplot as plt
+import itertools
+import time
 
 
+def load_secure(path: str):
+    while True:
+        try:
+            module = torch.load(path)
+            break
+        except:
+            time.sleep(0.1)
+    return module
+
 class Module(ABC, nn.Module):
     def __init__(self):
         super().__init__()
@@ -60,23 +71,40 @@ class Setup:
                  lrrt_max_decays: int = 0,
                  lrrt_decay: float = 0.9,
                  lrrt_initial_candidates: np.ndarray = np.array([1e-3, 1e-4, 1e-6]),
+                 overkill_initial_lr: float = 0.005,
+                 overkill_decay: float = 0.9,
+                 overkill_max_violations: int = None,
+                 overkill_max_decays: int = 10,
                  es_max_violations: int = 2,
                  optimizer_class=torch.optim.Adam):
         self.loader_train = loader_train
         self.loader_val = loader_val
         self.loader_test = loader_test
         self.device = device
+        self.optimizer_class = optimizer_class
         self.monitor_n_losses = monitor_n_losses  # prints loss slope after this amount of training steps
         self.checkpoint_initial = checkpoint_initial
         self.checkpoint_running = checkpoint_running
         self.checkpoint_final = checkpoint_final
+
+        # lrrt
         self.lrrt_n_batches = lrrt_n_batches  # batches used in lrrt for learning rate determination
         self.lrrt_slope_desired = lrrt_slope_desired  # exclusive border
         self.lrrt_max_decays = lrrt_max_decays  # max number of candidate decays performed in lrrt
         self.lrrt_decay = lrrt_decay
         self.lrrt_initial_candidates = lrrt_initial_candidates
-        self.es_max_violations = es_max_violations  # max number of early stopping violations
-        self.optimizer_class = optimizer_class
+
+        # overkill
+        self.overkill_initial_lr = overkill_initial_lr
+        self.overkill_decay = overkill_decay
+        if overkill_max_violations is None:
+            self.overkill_max_violations = len(self.loader_train)
+        else:
+            self.overkill_max_violations = overkill_max_violations
+        self.overkill_max_decays = overkill_max_decays
+
+        # early stopping
+        self.es_max_violations = es_max_violations
 
 
 class BatchHandler:
@@ -111,7 +139,10 @@ class BatchHandler:
         scores = self.forward_batch(module=module, batch=batch)["scores"]
         return torch.argmax(scores, dim=1)
 
-    def train_n_batches(self, module: Module, optimizer: Optimizer, n_batches: int, loader: DataLoader,
+    def train_n_batches(self, module: Module,
+                        optimizer: Optimizer,
+                        n_batches: int,
+                        loader: DataLoader,
                         freeze_pretrained: bool):
         losses = []
         for train_iter, batch in enumerate(loader):
@@ -149,7 +180,7 @@ class BatchHandler:
         for decay_it in range(self.setup.lrrt_max_decays + 1):
             candidate_slopes = np.zeros(shape=len(candidate_lrs))
             for i, lr_candidate in enumerate(candidate_lrs):
-                module = torch.load(self.setup.checkpoint_running)
+                module = load_secure(self.setup.checkpoint_running)
                 optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr_candidate)
                 candidate_slopes[i] = self.train_n_batches(module=module,
                                                            optimizer=optimizer,
@@ -177,12 +208,12 @@ class Debugger(BatchHandler):
     def __init__(self, setup: Setup):
         super(Debugger, self).__init__(setup=setup)
 
-    def overfit_one_batch(self, module: Module,
-                          batch_debug: list,
-                          n_iters: int,
-                          lrrt_loader: DataLoader = None,
-                          lr: float = 1e-5,
-                          freeze_pretrained: bool = False):
+    def overfit_batch(self, module: Module,
+                      batch_debug: list,
+                      n_iters: int,
+                      lrrt_loader: DataLoader = None,
+                      lr: float = 1e-5,
+                      freeze_pretrained: bool = False):
         if lrrt_loader is not None:
             lr, _ = self.lrrt(loader=lrrt_loader, freeze_pretrained=freeze_pretrained)
         optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr)
@@ -208,49 +239,52 @@ class Visualizer:
             i -= 1
         return 1
 
-    def plot_lines(self, lines: list, title: str, subplot_titles: list, y_label: str, x_label: str, file_name: str):
+    def lines_multiplot(self, lines: list, title: str, multiplot_titles: list, y_label: str, x_label: str, file_name: str):
+        """
+        creates multiple lines in the same subplot
+        :param lines:
+        :param title:
+        :param multiplot_titles:
+        :param y_label:
+        :param x_label:
+        :param file_name:
+        :return:
+        """
+        plt.figure(figsize=(4, 4))
+        for i, line in enumerate(lines):
+            plt.plot(range(len(line)), line, label=multiplot_titles[i])
+        plt.ylabel = y_label
+        plt.xlabel = x_label
+        plt.legend()
+        plt.tight_layout()
+        plt.savefig(f"../monitoring/{file_name}.png")
+
+    def lines_subplot(self, lines: list, title: str, subplot_titles: list, y_label: str, x_label: str, file_name: str):
+        """
+        creates multiple lines in individual subplots
+
+        :param lines:
+        :param title:
+        :param subplot_titles:
+        :param y_label:
+        :param x_label:
+        :param file_name:
+        :return:
+        """
         n_lines = len(lines)
-        n_rows = self.largest_divisor(n=n_lines)
-        n_cols = n_lines // n_rows
-        plt.figure(figsize=(20, 10))
+        n_cols = self.largest_divisor(n=n_lines)
+        n_rows = n_lines // n_cols
+        plt.figure(figsize=(n_cols*4, n_rows*4))
         plt.suptitle(title)
         for i in range(n_lines):
             plt.subplot(n_rows, n_cols, i + 1)
             plt.title(subplot_titles[i])
-            plt.ylabel(y_label)
-            plt.xlabel(x_label)
+            plt.ylabel = y_label
+            plt.xlabel = x_label
             plt.plot(range(len(lines[i])), lines[i])
         plt.tight_layout()
         plt.savefig(f"../monitoring/{file_name}.png")
 
-    def debug_lr(self, n_iters: int, batch_debug: list,
-                 lr_candidates: list = [5e-1, 1e-1, 5e-3, 1e-3, 5e-4, 1e-4, 5e-5, 1e-5, 5e-6, 1e-6],
-                 fig_name: str = "lr_debug"):
-        """
-        :param lr_candidates: len(lr_candidates) % 2 == 0
-        :param n_iters:
-        :param batch_debug:
-        :param fig_name:
-        :return:
-        """
-        plt.figure(figsize=(20, 10))
-        plt.suptitle("overfitting one batch with various learning rates")
-        for i, cand in enumerate(lr_candidates):
-            mimo = torch.load(self.setup.checkpoint_initial)
-            adam = torch.optim.Adam(params=mimo.parameters(), lr=cand)
-            _, losses_debug = self.overfit_one_train_batch(module=mimo,
-                                                           batch=batch_debug,
-                                                           optimizer=adam,
-                                                           n_iters=n_iters,
-                                                           freeze_pretrained=False)
-            plt.subplot(2, int(len(lr_candidates) / 2), i + 1)
-            plt.title("lr: " + str(cand))
-            plt.ylabel("loss")
-            plt.xlabel("training iteration")
-            plt.plot(torch.arange(len(losses_debug)), losses_debug)
-        plt.tight_layout()
-        plt.savefig(f"../monitoring/{fig_name}.png")
-
 
 class Trainer(BatchHandler):
     def __init__(self, setup: Setup):
@@ -267,54 +301,51 @@ class Trainer(BatchHandler):
         loss_epoch_val = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val)
         return loss_epoch_train, loss_epoch_val
 
-    def train_n_epochs_initial_lrrt(self, n_epochs: int, freeze_pretrained: bool = False):
-        """
-        determines the initial learning rate per epoch using lrrt.
-
-        :param int n_epochs: #training epochs after determining the initial learning rate with lrrt
-        :param bool freeze_pretrained:
-        :return: trained module
-        """
-        module = torch.load(self.setup.checkpoint_running)
-        loss_train, loss_val_last = self.losses_epoch_eval(module=module)
-        print("initial eval loss val", loss_val_last, "initial eval loss train", loss_train)
-        for epoch in range(1, n_epochs + 1):
-            print("training epoch", epoch)
-            lr_best, _ = self.lrrt(freeze_pretrained=freeze_pretrained, loader=self.setup.loader_train)
-            module = torch.load(self.setup.checkpoint_running).to(self.setup.device)
-            optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr_best)
-            self.train_n_batches(module=module, optimizer=optimizer, n_batches=len(self.setup.loader_train),
-                                 freeze_pretrained=freeze_pretrained, loader=self.setup.loader_train)
-
-            loss_train, loss_val = self.losses_epoch_eval(module=module)
-            print("eval loss val", loss_val, "eval loss train", loss_train)
-            torch.save(module, self.setup.checkpoint_running)
-            torch.save(module, self.setup.checkpoint_final)
-        return torch.load(self.setup.checkpoint_final)
-
-    def train_n_epochs_early_stop_initial_lrrt(self, max_epochs: int, freeze_pretrained_layers: bool):
+    def train(self, max_epochs: int, freeze_pretrained: bool = False):
         """
         determines the initial learning rate per epoch using lrrt.
         early stops (naively after one early stop violation)
 
         :param int max_epochs: max #training epochs after determining the initial learning rate with lrrt
-        :param bool freeze_pretrained_layers:
-        :return: early stopped trained module
+        :param bool freeze_pretrained:
+        :return: early stopped trained module, lrrt chosen learning rates, epoch train losses, epoch val losses
         """
         es_violations = 0
-        module = torch.load(self.setup.checkpoint_running)
+        losses_train = []
+        losses_val = []
+        best_lrs = []
+
+        module = load_secure(self.setup.checkpoint_running)
         loss_train, loss_val_last = self.losses_epoch_eval(module=module)
         print("initial eval loss val", loss_val_last, "initial eval loss train", loss_train)
+        losses_train.append(loss_train)
+        losses_val.append(loss_val_last)
+        best_lrs.append(0)
+
         for epoch in range(1, max_epochs + 1):
             print("training epoch", epoch)
-            lr_best, _ = self.lrrt(freeze_pretrained=freeze_pretrained_layers, loader=self.setup.loader_train)
-            module = torch.load(self.setup.checkpoint_running).to(self.setup.device)
-            optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr_best)
-            self.train_n_batches(module=module, optimizer=optimizer, n_batches=len(self.setup.loader_train),
-                                 freeze_pretrained=freeze_pretrained_layers, loader=self.setup.loader_train)
 
+            # lrrt
+            lr_best, _ = self.lrrt(freeze_pretrained=freeze_pretrained,
+                                   loader=self.setup.loader_train)
+
+            # epoch training
+            module = load_secure(self.setup.checkpoint_running).to(self.setup.device)
+            optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr_best)
+            self.train_n_batches(module=module,
+                                 optimizer=optimizer,
+                                 n_batches=len(self.setup.loader_train),
+                                 freeze_pretrained=freeze_pretrained,
+                                 loader=self.setup.loader_train)
+
+            # epoch evaluation
             loss_train, loss_val = self.losses_epoch_eval(module=module)
             print("eval loss val", loss_val, "eval loss train", loss_train)
+            losses_train.append(loss_train)
+            losses_val.append(loss_val)
+            best_lrs.append(lr_best)
+
+            # early stopping checkpointing
             if loss_val < loss_val_last:
                 torch.save(module, self.setup.checkpoint_running)
                 torch.save(module, self.setup.checkpoint_final)
@@ -328,4 +359,51 @@ class Trainer(BatchHandler):
                 if es_violations == self.setup.es_max_violations:
                     print("early stopping")
                     break
-        return torch.load(self.setup.checkpoint_final)
+        return load_secure(self.setup.checkpoint_final), best_lrs, losses_train, losses_val
+
+    def train_overkill_epoch(self, lr_initial: float, freeze_pretrained: bool):
+        violation_counter = len(self.setup.loader_train)
+        best_lrs = []
+        module = load_secure(self.setup.checkpoint_running)
+        loss_val_last = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val)
+        for i, batch in enumerate(self.setup.loader_train):
+            print("iter", i+1, "of", len(self.setup.loader_train))
+            lr = lr_initial
+            for i in range(self.setup.overkill_max_decays + 1):
+                module = load_secure(self.setup.checkpoint_running)
+                optimizer = self.setup.optimizer_class(params=module.parameters(), lr=lr)
+                self.train_batch(module=module, optimizer=optimizer, batch=batch, freeze_pretrained=freeze_pretrained)
+                loss = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val)
+                if loss < loss_val_last:
+                    print("improvement, loss", loss)
+                    loss_val_last = loss
+                    torch.save(module, self.setup.checkpoint_running)
+                    best_lrs.append(lr)
+                    violation_counter -= 1
+                    break
+                lr = lr * self.setup.overkill_decay
+        return best_lrs, loss, violation_counter
+
+    def train_overkill(self, max_epochs: int, freeze_pretrained: bool = False):
+        losses_train = []
+        losses_val = []
+        best_lrs = []
+        lr = self.setup.overkill_initial_lr
+        module = load_secure(self.setup.checkpoint_running)
+        losses_train.append(self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_train))
+        losses_val.append(self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val))
+        for epoch in range(max_epochs):
+            print("epoch", epoch + 1)
+            best_lrs_epoch, loss_val_epoch, violations = self.train_overkill_epoch(freeze_pretrained=freeze_pretrained, lr_initial=lr)
+            best_lrs.append(best_lrs_epoch)
+            losses_val.append(loss_val_epoch)
+            module = load_secure(self.setup.checkpoint_running)
+            losses_train.append(self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_train))
+            print("train loss:", losses_train[-1])
+            print("val loss:", losses_val[-1])
+            lr = float(torch.tensor(best_lrs[-1]).mean())
+            if violations >= self.setup.overkill_max_violations:
+                print("too few updates in this epoch, convergence")
+                break
+        torch.save(module, self.setup.checkpoint_final)
+        return module, list(itertools.chain(*best_lrs)), losses_train, losses_val
